package com.tzld.piaoquan.ad.engine.service.score.strategy;

import com.alibaba.fastjson.JSON;
import com.ctrip.framework.apollo.spring.annotation.ApolloJsonValue;
import com.tzld.piaoquan.ad.engine.commons.dto.AdPlatformCreativeDTO;
import com.tzld.piaoquan.ad.engine.commons.param.RankRecommendRequestParam;
import com.tzld.piaoquan.ad.engine.commons.score.ScoreParam;
import com.tzld.piaoquan.ad.engine.commons.util.WeightRandom;
import com.tzld.piaoquan.ad.engine.service.feature.Feature;
import com.tzld.piaoquan.ad.engine.service.feature.FeatureService;
import com.tzld.piaoquan.recommend.feature.domain.ad.base.AdRankItem;
import lombok.AllArgsConstructor;
import lombok.Data;
import lombok.NoArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import java.util.*;
import java.util.stream.Collectors;

@Slf4j
@Service
public class RankStrategyBy687 extends RankStrategyBasic {

    @Autowired
    private RankStrategyBy688 rankStrategyBy688;

    @ApolloJsonValue("${cid.open.hot.rank.user.layer :[]}")
    private Set<String> openHotRankUserLayers;

    @Value("${cid.hot.rank.view.threshold:5000}")
    private Integer hotRankViewThreshold;

    @Value("${cid.hot.rank.calc.weight.coefficient:2.5}")
    private Double hotRankCalcWeightCoefficient;

    @Value("${cid.hot.rank.max.weight:1000}")
    private Double hotRankMaxWeight;
    @Autowired
    private FeatureService featureService;

    @Override
    public List<AdRankItem> adItemRank(RankRecommendRequestParam request, ScoreParam scoreParam) {
        // 没有mid或者未配置分层，走基线排序
        if (StringUtils.isEmpty(request.getMid()) && CollectionUtils.isEmpty(openHotRankUserLayers)) {
            return rankStrategyBy688.adItemRank(request, scoreParam);
        }

        // 当前分层未开启此策略
        String userLayer = this.getUserLayer(request.getMid());
        if (!openHotRankUserLayers.contains(userLayer)) {
            return rankStrategyBy688.adItemRank(request, scoreParam);
        }

        List<AdPlatformCreativeDTO> recallCreativeList = request.getAdIdList();

        // 获取所有创意对应的客户列表
        List<Long> creativeIds = recallCreativeList.stream().map(AdPlatformCreativeDTO::getCreativeId).collect(Collectors.toList());
        Feature feature = featureService.getCreativeBasicInfo(creativeIds);
        this.fullCustomerInfo(feature.getCidFeature(), recallCreativeList);

        List<String> customers = recallCreativeList.stream()
                .map(AdPlatformCreativeDTO::getCustomer)
                .distinct()
                .collect(Collectors.toList());

        // 获取客户列表，并选择一个客户
        String keyFormat = "ad:engine:customer:layer:info:%s:" + userLayer;
        List<HotRankFeatureInfo> customerFeature = this.multiGetFeature(customers, keyFormat);
        String customer = this.choose(customerFeature);

        // 从当前客户的所有创意中，选择一个创意
        List<String> customerCreativeIds = recallCreativeList.stream()
                .filter(c -> StringUtils.equals(c.getCustomer(), customer))
                .map(AdPlatformCreativeDTO::getCreativeId)
                .map(Objects::toString)
                .collect(Collectors.toList());
        String cidKeyFormat = "ad:engine:cid:layer:info:%s:" + userLayer;
        List<HotRankFeatureInfo> creativeFeature = this.multiGetFeature(customerCreativeIds, cidKeyFormat);
        String chooseCreativeId = this.choose(creativeFeature);

        return null;
    }

    private String getUserLayer(String mid) {

        String key = String.format("ad:engine:mid:layer:%s", mid);
        String value = algRedisHelper.get(key);
        if (StringUtils.isEmpty(value)) {
            return "无曝光";
        }
        return value;
    }

    private List<HotRankFeatureInfo> multiGetFeature(List<String> customers, String keyFormat) {
        List<HotRankFeatureInfo> hotRankFeatureInfos = new ArrayList<>(customers.size());

        List<String> keys = new ArrayList<>(customers.size());
        for (String customer : customers) {
            String key = String.format(keyFormat, customer);
            keys.add(key);
        }

        List<String> values = algRedisHelper.mget(keys);

        for (int i = 0; i < customers.size(); i++) {
            String value = values.get(i);
            HotRankFeatureInfo hotRankFeatureInfo = JSON.parseObject(value, HotRankFeatureInfo.class);
            hotRankFeatureInfo.setLabel(customers.get(i));
            hotRankFeatureInfos.add(hotRankFeatureInfo);
        }

        return hotRankFeatureInfos;
    }

    private void fullCustomerInfo(Map<String, Map<String, Map<String, String>>> cidFeature, List<AdPlatformCreativeDTO> dtoList) {
        for (AdPlatformCreativeDTO dto : dtoList) {
            Map<String, Map<String, String>> feature = cidFeature.getOrDefault(String.valueOf(dto.getCreativeId()), new HashMap<>());
            Map<String, String> basicInfo = feature.getOrDefault("alg_cid_feature_basic_info", new HashMap<>());
            String customer = basicInfo.getOrDefault("customer", "");
            dto.setCustomer(customer);
        }
    }

    private String choose(List<HotRankFeatureInfo> hotRankFeatureInfos) {
        // 按CPM排序
        hotRankFeatureInfos.sort(Comparator.comparing(HotRankFeatureInfo::getCpm));

        // 按照曝光拆分
        List<HotRankFeatureInfo> highView = new ArrayList<>();
        List<HotRankFeatureInfo> tailView = new ArrayList<>();
        for (HotRankFeatureInfo hotRankFeatureInfo : hotRankFeatureInfos) {
            if (hotRankFeatureInfo.getView() >= hotRankViewThreshold) {
                highView.add(hotRankFeatureInfo);
            } else {
                tailView.add(hotRankFeatureInfo);
            }
        }

        if (CollectionUtils.isNotEmpty(highView)) {
            // 高曝光中CPM最大的设置初始权重
            this.calcFirstItemWeight(highView, tailView, true);
            // 计算高曝光列表中其他的权重
            this.calcListOtherWeight(highView);
        }

        if (CollectionUtils.isNotEmpty(tailView)) {
            this.calcFirstItemWeight(highView, tailView, false);
            this.calcListOtherWeight(tailView);
        }

        List<WeightRandom.ItemWithWeight<String>> itemWithWeights = new ArrayList<>(hotRankFeatureInfos.size());
        for (HotRankFeatureInfo hotRankFeatureInfo : hotRankFeatureInfos) {
            WeightRandom.ItemWithWeight<String> itemWeight = new WeightRandom.ItemWithWeight<>(hotRankFeatureInfo.getLabel(), hotRankFeatureInfo.getWeight());
            itemWithWeights.add(itemWeight);
        }

        return new WeightRandom<>(itemWithWeights).choose();
    }

    private void calcFirstItemWeight(List<HotRankFeatureInfo> high, List<HotRankFeatureInfo> tail, boolean isHigh) {
        if (isHigh) {
            high.get(0).setWeight(Math.max(1, hotRankMaxWeight));
        } else {
            HotRankFeatureInfo firstTail = tail.get(0);
            double firstCpm = firstTail.getCpm();
            for (HotRankFeatureInfo info : high) {
                if (firstCpm >= info.getCpm()) {
                    // 权重最小为1
                    double weight = Math.max(1, info.getWeight() / hotRankCalcWeightCoefficient);
                    firstTail.setWeight(weight);
                    return;
                }
            }
            firstTail.setWeight(1);
        }
    }

    /**
     * 计算列表中第二个元素往后的权重
     */
    private void calcListOtherWeight(List<HotRankFeatureInfo> rankItem) {
        for (int i = 1; i < rankItem.size(); i++) {
            HotRankFeatureInfo curr = rankItem.get(i);
            if (curr.getCpm() == 0) {
                curr.setWeight(1);
                continue;
            }
            HotRankFeatureInfo prev = rankItem.get(i - 1);
            double currWeight1 = prev.getWeight() * (1 - (1 - curr.getCpm() / prev.getCpm())) * hotRankCalcWeightCoefficient;
            double currWeight2 = prev.getWeight() / hotRankCalcWeightCoefficient;
            double currWeight = Math.max(currWeight1, currWeight2);
            curr.setWeight(Math.max(1, currWeight));
        }
    }

    @Data
    @NoArgsConstructor
    @AllArgsConstructor
    private static class HotRankFeatureInfo {
        private String label;

        private double cpm;

        private int view;

        private double weight = 1;
    }
}
